import subprocess
import struct
import random
import os
import hashlib
import sys

def _hash(x):
  return hashlib.sha256(repr(x).encode()).hexdigest()[:8]
def print_hash(name, x):
  print("{:<32}: {}".format(name, _hash(x)))

def parse_binary(b):
  return ''.join(map(lambda x:format(int(x), '0>2x'), b))

def print_header(s):
  l = 80
  print('{}\n{:^80}\n{}\n'.format(l * '_', s, l * '_'))

def parse_frame(buf):
  c = [0]
  def b(c, n):
    ret = buf[c[0]:c[0]+n]
    c[0] += n
    return ret
  dst_mac             = b(c, 6)
  src_mac             = b(c, 6)
  ether_type          = b(c, 2)
  ecpri_common_header = b(c, 4)
  pc_id               = b(c, 2)
  seq_id              = b(c, 2)
  iq_samples          = b(c, 240)

  avg = sum(map(lambda x:int(x), iq_samples))
  s = '''
| {:^12} | {:^12} | {:^12} | {:^12} | {:^12} | {:^12} |
| {:^12} | {:^12} | {:^12} | {:^12} | {:^12} | {:^12} |
| {} |'''.format(
              "DST", "SRC", "ETHER TYPE", "ECPRI HEADER", "PC ID", "SEQ ID",
              parse_binary(dst_mac),
              parse_binary(src_mac),
              parse_binary(ether_type),
              parse_binary(ecpri_common_header),
              parse_binary(pc_id),
              parse_binary(seq_id),
              parse_binary(iq_samples))[1:]
  
  return (avg == 0, s)

def read_trace(trace_dir, name, n):
  file_name = '{}/{}'.format(trace_dir, name)

  file_size = os.path.getsize(file_name)
  n = min(n * 262, file_size)
  f = open(file_name, "rb")
  data = f.read(n)
  f.close()

  print(name + ": {} frames, reading {} frames".format(file_size / 262, n / 262))

  return data

def print_frame(data):
  n = len(data)
  frame_len = 262

  for i in range(int(n / frame_len)):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    if empty:
      continue
    print("Frame example:\n\n{}\n".format(s))
    break

def analyze_tdd(data):
  null_frames = 0
  tdd_switch_list = []
  prev_empty = False
  prev_i = 0
  n = len(data)
  frame_len = 262

  i = 0
  for i in range(int(n / frame_len)):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    if i > 0 and empty != prev_empty:
        tdd_switch_list.append(i-prev_i)
        prev_i = i
    null_frames += empty
    prev_empty = empty
  tdd_switch_list.append(i-prev_i)

  print('Frame analyzed: ' + str(i))

  total_frames = int(n / frame_len)
  ratio = total_frames / null_frames if null_frames > 0 else 'inf'
  print('TDD ratio: {}'.format(ratio))
  print('TDD switch list: ' + ', '.join(map(str, tdd_switch_list)) + '\n')

def print_frame_list(data, start, end):
  n = len(data)
  frame_len = 262

  end = int(n / frame_len) if end == -1 else end
  end = min(int(n / frame_len), end)

  for i in range(start, end):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    print(s)
    print('')

def print_iq_list(data, start, end, n_channel=4):
  n = len(data)
  frame_len = 64 * n_channel

  end = int(n / frame_len) if end == -1 else end
  end = min(int(n / frame_len), end)

  for i in range(start * 4, end * 4, 4):
    iq = data[i * frame_len:(i+1) * frame_len]
    iq_packed = []
    k = 0

    h = _hash(iq)
    prev_x = 2
    for x in map(lambda x: int(x != 0), iq):
      if x == prev_x:
        iq_packed[-1][0] += 1
      else:
        iq_packed.append([1, str(x) * 2])
      prev_x = x
    print(h + "  " + " ".join(map(lambda x: "{}*{}".format(*x), iq_packed)))

def read_trx_trace(trace_dir, name, n, channels):
  file_name = '{}/{}'.format(trace_dir, name)

  file_size = os.path.getsize(file_name)
  n = min(n * (4 * channels * 64), file_size)
  f = open(file_name, "rb")
  data = f.read(n)
  f.close()

  print(name + ": {} frames, reading {} frames".format(file_size / (4 * channels * 64), n / (4 * channels * 64)))

  return data

def analyze_trx_tdd(data, channels):
  null_frames = 0
  tdd_switch_list = []
  prev_empty = False
  prev_i = 0
  n = len(data)
  frame_len = channels * 4 * 64

  max_iq_sample = 0
  min_iq_sample = 0
  avg_iq_sample = 0
  
  i = 0
  for i in range(int(n / frame_len)):
    iq_samples = data[i * frame_len:(i+1) * frame_len]
    iq_samples = [struct.unpack('f', iq_samples[4*j:4*(j+1)])[0] for j in range(int(len(iq_samples) / 4))]

    iq_samples_abs_avg = sum(map(abs, iq_samples))
    max_iq_sample = max(max_iq_sample, max(iq_samples))
    min_iq_sample = min(min_iq_sample, min(iq_samples))
    avg_iq_sample += iq_samples_abs_avg
  
    empty = iq_samples_abs_avg < 0.01
    if i == 0:
      first_tdd_period = empty
    if i > 0 and empty != prev_empty:
        tdd_switch_list.append(i-prev_i)
        prev_i = i
    null_frames += empty
    prev_empty = empty
  tdd_switch_list.append(i-prev_i)
  if n:
    avg_iq_sample /= (n / frame_len) * (64 * channels)

  print('Max IQ sample: {}, Min IQ sample: {}'.format(max_iq_sample, min_iq_sample))
  print('Avg IQ sample: {}'.format(avg_iq_sample))
  print('Frame analyzed: ' + str(i))

  total_frames = int(n / frame_len)
  ratio = total_frames / null_frames if null_frames > 0 else 'inf'
  print('TDD ratio: {}'.format(ratio))
  print('TDD switch list: ' + ', '.join(map(str, tdd_switch_list)) + '\n')
  print('First TDD period: ' + "Not emitting" if n and first_tdd_period else "emitting")

BF1_PATH="/home/slapos/trx-ecpri-testing/bf1/bf1"

def float_to_byte(float_list):
  iq_samples_bytes = bytearray(64 * 4)
  for i in range(64):
    b = struct.pack('f', float_list[i])
    for j in range(4):
      iq_samples_bytes[i * 4 + j] = b[j]
  return bytes(iq_samples_bytes)

def byte_to_float(b):
  float_list = []
  for i in range(64):
    x = struct.unpack('f', b[4 * i:4 * (i+1)])[0]
    float_list.append(x)
  return float_list

def compress(data):
  s = subprocess.run(
                  [BF1_PATH, "c"],
                  input=data,
                  capture_output=True,
  )
  return s.stdout

def decompress(data):
  s = subprocess.run([BF1_PATH, "d"],
                  input=data,
                  capture_output=True,
  )
  return s.stdout

def open_all(trace_dir, n):
  return (
    read_trace(trace_dir, 'tx.trace', n),
    read_trace(trace_dir, 'rx.trace', n),
    read_trx_trace(trace_dir, 'trxw.trace', n, 4),
    read_trx_trace(trace_dir, 'trxr.trace', n, 1),
  )

def analyze_all_tdd(trace_dir, n=0, data=None):

  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
    print_header("TX TDD analysis")
    analyze_tdd(tx_data)
    print_header("RX TDD analysis")
    analyze_tdd(rx_data)
    print_header("TRXW TDD analysis")
    analyze_trx_tdd(trxw_data, 4)
    print_header("TRXR TDD analysis")
    analyze_trx_tdd(trxr_data, 1)
  else:
    data = read_trace(trace_dir, 'rx.trace', n)
    analyze_tdd(data)
    data = read_trace(trace_dir, 'tx.trace', n)
    analyze_tdd(data)
    data = read_trx_trace(trace_dir, 'trxw.trace', n, 4)
    analyze_trx_tdd(data, 4)
    data = read_trx_trace(trace_dir, 'trxr.trace', n, 1)
    analyze_trx_tdd(data, 1)

def check_tx_trxw(trace_dir, n=0, data=None):
  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
  else:
    tx_data = read_trace(trace_dir, 'tx.trace', n)
    trxw_data = read_trx_trace('trxw.trace', n, 4)
  for i in range(int(min(len(tx_data) / 262, len(trxw_data) / (64 * 4 * 4)))):
    for j in range(1):
      k = i * 4 + j
      tx_iq = tx_data[(i * 262 + 22 + j * 60):(i * 262 + 22 + (j+1) * 60)]
      trxw_iq = trxw_data[(k * (64 * 4)):((k+1) * (64 * 4))]
      tx_iq_d = decompress(tx_iq)
      trxw_iq_d = compress(trxw_iq)
      if (_hash(tx_iq) != _hash(trxw_iq_d)) or (_hash(trxw_iq) != _hash(tx_iq_d)):
        print((i,j))
        print_hash("TX", tx_iq)
        print_hash("TRX compressed", trxw_iq_d)
        print_hash("TRX", trxw_iq)
        print_hash("TX decompressed", tx_iq_d)

def print_all(n=0, data=None):
  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
    print_header('TX DATA')
    print_frame_list(tx_data, 0, 10)
    print_header('RX DATA')
    print_header('TRXW DATA')
    print_iq_list(trxw_data, 0, 10, 4)
    print_header('TRXR DATA')
    print_iq_list(trxr_data, 0, 10, 1)

def check_iq_ref(float_list, k):
  ref_iq = [0.2806, 0.1765, 0.241, 0.315, 0.1102, 0.1687, 0.2796, 0.1007, 0.1532, 0.1416, 0.3251, 0.1648, 0.0943, 0.2117, 0.0819, 0.2903, 0.2501, 0.0474, 0.2062, 0.0929, 0.1164, 0.2628, 0.233, 0.0807, 0.0622, 0.1898, 0.0353, 0.0621, 0.2952, 0.3302, 0.0593, 0.0149, 0.1332, 0.1623, 0.2873, 0.2878, 0.2234, 0.2496, 0.1093, 0.1939, 0.0139, 0.2309, 0.0788, 0.0413, 0.032, 0.2459, 0.0053, 0.1913, 0.0128, 0.0753, 0.2993, 0.2474, 0.2497, 0.2482, 0.0296, 0.316, 0.0697, 0.0964, 0.1447, 0.1936, 0.2217, 0.2048, 0.1711, 0.3068, 0.2425, 0.2568, 0.1369, 0.2054, 0.2321, 0.0719, 0.2702, 0.0296, 0.2425, 0.1185, 0.0195, 0.0897, 0.2867, 0.2843, 0.2479, 0.1714, 0.2353, 0.1614, 0.121, 0.1995, 0.0721, 0.1928, 0.0733, 0.2868, 0.2046, 0.0172, 0.0027, 0.0572, 0.1601, 0.2939, 0.1993, 0.2842, 0.0202, 0.1897, 0.0031, 0.1861, 0.1305, 0.1519, 0.0283, 0.2113, 0.2693, 0.0196, 0.3043, 0.0369, 0.0326, 0.1655, 0.2832, 0.2635, 0.1708, 0.1424, 0.0898, 0.1463, 0.0282, 0.2291, 0.0158, 0.0665, 0.1167, 0.2735, 0.1775, 0.3135, 0.0885, 0.2902, 0.0881, 0.2111, 0.2576, 0.2711, 0.3058, 0.1956, 0.1731, 0.0076, 0.0491, 0.2363, 0.2494, 0.2543, 0.2979, 0.3211, 0.0391, 0.2565, 0.2005, 0.3278, 0.3223, 0.2468, 0.1193, 0.2862, 0.0784, 0.1754, 0.1152, 0.2882, 0.0257, 0.079, 0.1545, 0.1653, 0.1096, 0.2328, 0.0746, 0.1112, 0.1772, 0.0905, 0.1404, 0.2764, 0.187, 0.138, 0.2594, 0.0333, 0.2169, 0.0259, 0.1605, 0.0934, 0.3246, 0.3075, 0.1163, 0.1556, 0.1923, 0.0994, 0.1196, 0.0013, 0.1475, 0.3157, 0.2043, 0.2966, 0.1123, 0.0212, 0.2995, 0.2666, 0.1253, 0.1169, 0.1842, 0.2481, 0.0942, 0.1586, 0.0143, 0.2794, 0.1392, 0.2289, 0.1335, 0.1818, 0.302, 0.2306, 0.1953, 0.1039, 0.2353, 0.2674, 0.1898, 0.0835, 0.0386, 0.1237, 0.1454, 0.019, 0.1964, 0.248, 0.1655, 0.1459, 0.2217, 0.1022, 0.1079, 0.0025, 0.0447, 0.1059, 0.1805, 0.1608, 0.2823, 0.078, 0.1647, 0.2305, 0.0922, 0.3287, 0.0871, 0.2898, 0.3061, 0.0612, 0.046, 0.215, 0.2707, 0.3313, 0.0825, 0.2921, 0.2113, 0.2864, 0.296, 0.1408, 0.2583, 0.1971, 0.0467, 0.1458, 0.2265, 0.3308, 0.1136, 0.0656, 0.0612, 0.2644, 0.2627, 0.1968]

  ref_iq2 = [ref_iq[(i + k) % len(ref_iq)] for i in range(len(float_list))]

  error_list = [abs(float_list[i] - ref_iq2[i]) for i in range(0, len(float_list))]
  error = max(error_list)
  if error > 0.1:
    print("Error (k = {}): {}".format(k, error))
  if error > 0.1:
    print('\n'.join(map(str, ["{}{} - {}: {} ({})".format("" if error_list[i] < 0.01 else "  ", i, error_list[i], float_list[i], ref_iq2[i]) for i in range(len(float_list))])))
  #print("{}, {}, {}".format(k, float_list[0], "Error" if error > 0.1 else "OK"))

def check_frame(data, n_channel, trace_offset=0):


  n = len(data)
  frame_len = 262

  start = 0
  end = int(n / frame_len)
  end = min(int(n / frame_len), end)

  k = trace_offset

  for i in range(start, end):
    frame = data[i * frame_len:(i+1) * frame_len]
    float_list = []
    for i in range(n_channel):
      iq_byte = decompress(frame[22 + i * 60:22 + (i+1) * 60])
      float_list += byte_to_float(iq_byte)
    check_iq_ref(float_list, k)
    k = (k + 1)

def check_iq(data, n_channel, trace_offset=0):
  
  n = len(data)
  frame_len = 64 * n_channel * 4

  start = 0
  end = int(n / frame_len)

  k = trace_offset

  for i in range(start, end):
    iq = data[i * frame_len:(i+1) * frame_len]
    float_list = []
    for j in range(n_channel):
      b = iq[256 * j : 256 * (j+1)]
      #b = decompress(compress(b))
      float_list += byte_to_float(b)
    check_iq_ref(float_list, k)
    k = (k + 1)

def check(data=None, rx=True, trace_offset=0):

  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
  else:
    return
  if rx:
    check_iq(trxr_data, 4, trace_offset=trace_offset)
    check_frame(rx_data, 4, trace_offset=trace_offset)
  else:
    check_iq(trxw_data, 4, trace_offset=trace_offset)
    check_frame(tx_data, 4, trace_offset=trace_offset)
  
if len(sys.argv) < 2:
  trace_dir = '.'
else:
  trace_dir = sys.argv[1]

trace_offset = 0
rx = True
if len(sys.argv) >= 3:
  trace_offset = int(sys.argv[2])
if len(sys.argv) >= 4:
  rx = sys.argv[3] == "rx"

data = open_all(trace_dir, 1000)
#check_tx_trxw(trace_dir, data=data)
#print_all(data=data)
#analyze_all_tdd(trace_dir, data=data)
check(data=data, rx=rx, trace_offset=trace_offset)
